from os import path
from subprocess import check_output

from setuptools import setup, find_packages

git_version_command = 'git describe --tags --long --dirty'
here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

if path.exists(here + "/.git"):
    init_file = None
    with open(path.join(here, 'jivago/__init__.py'), 'r', encoding='utf-8') as f:
        init_file = f.read()
    if "@@VERSION@@" in init_file:
        with open(path.join(here, 'jivago/__init__.py'), 'w', encoding='utf-8') as f:
            f.write(init_file.replace("@@VERSION@@",
                                      check_output(git_version_command.split()).decode('utf-8').strip().split("-")[0]))

setup(
    name='jivago',
    version_format='{tag}',
    setup_requires=['setuptools-git-version'],
    description='The highly-reflective object-oriented Python web framework',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/keotl/jivago',

    author='Kento A. Lauzon',
    author_email='kento.lauzon@ligature.ca',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],

    packages=find_packages(
        exclude=['contrib', 'docs', 'test', 'test.*' 'tests', 'tests.*', 'example_app', 'example_app.*', 'test_data',
                 'test_data.*', 'test_utils', 'test_utils.*', 'e2e_test', 'e2e_test.*']),

    install_requires=['Jinja2', 'MarkupSafe', 'croniter', 'pyyaml', 'werkzeug'],
)



import unittest
from unittest import mock

from jivago.event.config.annotations import EventHandler
from jivago.event.dispatch.message_dispatcher_runnable import MessageDispatcherRunnable
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Override, Inject
from jivago.lang.runnable import Runnable


class MessageDispatcherRunnableTest(unittest.TestCase):

    def setUp(self):
        self.service_locator = ServiceLocator()
        self.dependency_mock = mock.MagicMock()
        self.service_locator.bind(object, self.dependency_mock)
        self.service_locator.bind(MessageDispatcherRunnableMock, MessageDispatcherRunnableMock)

    def test_whenHandlingMessage_thenInstantiateAndCallRunMethod(self):
        dispatcher = MessageDispatcherRunnable("foo", MessageDispatcherRunnableMock, self.service_locator)

        dispatcher.handle(None)

        self.dependency_mock.assert_called()


@EventHandler("foo")
class MessageDispatcherRunnableMock(Runnable):

    @Inject
    def __init__(self, dependency_mock: object):
        self.dependency_mock = dependency_mock

    @Override
    def run(self):
        self.dependency_mock()



import unittest
from unittest import mock

from jivago.lang.registry import Registry
from jivago.serialization.serializer import Serializer
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.filter.system_filters.body_serialization_filter import BodySerializationFilter
from test_utils.request_builder import RequestBuilder
from test_utils.response_builder import ResponseBuilder

TEST_DATA = "foobar"
AN_UNSERIALIZABLE_OBJECT = object()

A_REQUEST = RequestBuilder().build()
A_DTO = {"foo": "bar"}


class BodySerializationFilterTest(unittest.TestCase):

    def setUp(self):
        self.serializerMock: Serializer = mock.create_autospec(Serializer)
        self.registryMock: Registry = mock.create_autospec(Registry)
        self.filter = BodySerializationFilter(self.serializerMock, self.registryMock)
        self.registryMock.is_annotated.return_value = True

        self.filterChainMock: FilterChain = mock.create_autospec(FilterChain)

    def test_givenSerializableBody_whenApplyingFilter_thenDtoIsConvertedToDictionaryByHandler(self):
        expected_body = {"data": TEST_DATA}
        response = ResponseBuilder().body(A_DTO).build()
        self.serializerMock.serialize.return_value = expected_body

        self.filter.doFilter(A_REQUEST, response, self.filterChainMock)

        self.assertEqual(expected_body, response.body)

    def test_givenDtoNotAnnotatedAsSerializable_whenApplyingFilter_thenBodyIsNotSerialized(self):
        response = ResponseBuilder().body(AN_UNSERIALIZABLE_OBJECT).build()
        self.registryMock.is_annotated.return_value = False

        self.filter.doFilter(A_REQUEST, response, self.filterChainMock)

        self.assertEqual(AN_UNSERIALIZABLE_OBJECT, response.body)



from jinja2 import Template

from jivago.lang.annotations import Override, Inject
from jivago.serialization.deserializer import Deserializer
from jivago.serialization.serializer import Serializer
from jivago.templating.rendered_view import RenderedView
from jivago.templating.view_template_repository import ViewTemplateRepository
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class TemplateFilter(Filter):

    @Inject
    def __init__(self, view_template_repository: ViewTemplateRepository,
                 serializer: Serializer, deserializer: Deserializer):
        self.deserializer = deserializer
        self.serializer = serializer
        self.view_template_repository = view_template_repository

    @Override
    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        chain.doFilter(request, response)

        if isinstance(response.body, RenderedView):
            rendered_view = response.body
            template_text = self.view_template_repository.get_template(rendered_view.view_file)
            template_parameters = rendered_view.data
            if self.deserializer.is_deserializable_type(rendered_view.data.__class__):
                template_parameters = self.serializer.serialize(rendered_view.data)
            response.body = Template(template_text).render(**template_parameters)

            response.headers['Content-Type'] = "text/html"



from datetime import datetime, date

from jivago.lang.annotations import Override
from jivago.serialization.serialization_strategy import SerializationStrategy


class DatetimeSerializationStrategy(SerializationStrategy):

    @Override
    def can_handle_serialization(self, obj) -> bool:
        return isinstance(obj, datetime) or isinstance(obj, date)

    @Override
    def serialize(self, obj) -> object:
        return str(obj)



from jivago.wsgi.request.response import Response


class ExceptionMapper(object):

    def handles(self, exception: Exception) -> bool:
        raise NotImplementedError

    def create_response(self, exception: Exception) -> Response:
        raise NotImplementedError



import yaml

from jivago.config.properties.application_config_loader import ApplicationConfigLoader
from jivago.config.properties.application_properties import ApplicationProperties
from jivago.lang.annotations import Override


class YamlConfigLoader(ApplicationConfigLoader):

    @Override
    def matches(self, filepath: str) -> bool:
        return filepath.endswith(".yml")

    @Override
    def read(self, filepath: str) -> ApplicationProperties:
        with open(filepath, 'r') as f:
            return ApplicationProperties(yaml.safe_load(f))



from datetime import datetime, timedelta

from jivago.lang.annotations import Override
from jivago.scheduling.annotations import _Interval
from jivago.scheduling.schedule import Schedule


class RegularIntervalSchedule(Schedule):

    def __init__(self, time_interval: _Interval, start_time: datetime = None):
        self.interval = time_interval.interval_time_in_seconds
        self.next_time = start_time if start_time else datetime.utcnow()

    def increment(self):
        self.next_time += timedelta(seconds=self.interval)

    @Override
    def next_start_time(self) -> datetime:
        self.increment()
        return self.next_time



import os
from typing import List

from jivago.lang.annotations import Override
from jivago.lang.stream import Stream
from jivago.wsgi.methods import GET
from jivago.wsgi.routing.route_registration import RouteRegistration
from jivago.wsgi.routing.routing_table import RoutingTable
from jivago.wsgi.routing.serving.static_file_serving_resource import StaticFileServingResource
from jivago.wsgi.routing.table.path_util import split_path


class StaticFileRoutingTable(RoutingTable):

    def __init__(self, folder_root: str, allowed_extensions: List[str] = None):
        """allowed_extensions = None to allow any file extension."""
        self.allowed_extensions = allowed_extensions
        self.folder_root = folder_root

    @Override
    def get_route_registrations(self, path: str) -> List[RouteRegistration]:
        filepath = os.path.join(self.folder_root, path.lstrip("/"))
        if not os.path.exists(filepath) or self.__is_disallowed_extension(path) or os.path.isdir(filepath):
            return []
        return [RouteRegistration(StaticFileServingResource(filepath),
                                  StaticFileServingResource.serve_file,
                                  split_path(path),
                                  GET)]

    def __is_disallowed_extension(self, path: str) -> bool:
        if self.allowed_extensions is None:
            return False
        return Stream(self.allowed_extensions).noneMatch(lambda extension: path.endswith(extension))



from jivago.config.router.router_config_rule import RouterConfigRule
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.registry import Registry
from jivago.serialization.deserializer import Deserializer
from jivago.wsgi.filter.filter_chain_factory import FilterChainFactory
from jivago.config.router.filtering.filtering_rule import FilteringRule
from jivago.wsgi.invocation.route_handler_factory import RouteHandlerFactory
from jivago.wsgi.request.request_factory import RequestFactory
from jivago.wsgi.routing.cors.cors_request_handler_factory import CorsRequestHandlerFactory
from jivago.config.router.cors_rule import CorsRule
from jivago.wsgi.routing.router import Router
from jivago.wsgi.routing.routing_rule import RoutingRule


class RouterBuilder(object):

    def __init__(self):
        self.filtering_rules = []
        self.routing_rules = []
        self.cors_rules = []

    def add_rule(self, rule: RouterConfigRule) -> "RouterBuilder":
        if isinstance(rule, FilteringRule):
            self.filtering_rules.append(rule)
        elif isinstance(rule, RoutingRule):
            self.routing_rules.append(rule)
        elif isinstance(rule, CorsRule):
            self.cors_rules.append(rule)
        return self

    def build(self, registry: Registry, service_locator: ServiceLocator) -> Router:
        filter_chain_factory = FilterChainFactory(self.filtering_rules, service_locator,
                                                  RouteHandlerFactory(service_locator,
                                                                      Deserializer(registry),
                                                                      self.routing_rules,
                                                                      CorsRequestHandlerFactory(self.cors_rules))
                                                  )
        return Router(service_locator, RequestFactory(), filter_chain_factory)



from jivago.lang.annotations import Override
from jivago.wsgi.invocation.missing_route_invocation_argument import MissingRouteInvocationArgument
from jivago.wsgi.invocation.parameter_selection.query_parameter_selector import QueryParameterSelector
from jivago.wsgi.invocation.parameters import OptionalQueryParam
from jivago.wsgi.request.request import Request


class OptionalQueryParameterSelector(QueryParameterSelector):

    @Override
    def matches(self, paramater_declaration: type) -> bool:
        return hasattr(paramater_declaration, '__origin__') and paramater_declaration.__origin__ == OptionalQueryParam

    @Override
    def format_parameter(self, parameter_name: str, parameter_declaration: type, request: Request) -> object:
        try:
            return super().format_parameter(parameter_name, parameter_declaration, request)
        except MissingRouteInvocationArgument:
            return None



import unittest

from jivago.wsgi.routing.cors.no_matching_cors_rule_exception import NoMatchingCorsRuleException
from jivago.wsgi.routing.cors.no_matching_cors_rule_exception_mapper import NoMatchingCorsRuleExceptionMapper
from test_utils.request_builder import RequestBuilder


class NoMatchingCorsRuleExceptionMapperTest(unittest.TestCase):

    def setUp(self):
        self.mapper = NoMatchingCorsRuleExceptionMapper()

    def test_matchesException(self):
        matches_exception = self.mapper.handles(NoMatchingCorsRuleException())

        self.assertTrue(matches_exception)

    def test_whenCreatingResponse_thenReturn400BadRequest(self):
        response = self.mapper.create_response(RequestBuilder().build())

        self.assertEqual(400, response.status)



from typing import List

from jivago.wsgi.methods import HttpMethod
from jivago.wsgi.routing.exception.unknown_path_exception import UnknownPathException
from jivago.wsgi.routing.route_registration import RouteRegistration

PATH_PARAMETER = object()


class RouteNode(object):
    def __init__(self):
        self.children = {}
        self.invocators = {}

    def register_child(self, path: List[str], http_primitive: HttpMethod, route_registration: RouteRegistration):
        if len(path) == 0:
            if http_primitive in self.invocators:
                self.invocators[http_primitive].append(route_registration)
            else:
                self.invocators[http_primitive] = [route_registration]
        else:
            next_path_element = PATH_PARAMETER if path[0].startswith("{") and path[0].endswith('}') else path[0]
            if next_path_element not in self.children:
                self.children[next_path_element] = RouteNode()
            self.children[next_path_element].register_child(path[1::], http_primitive, route_registration)

    def explore(self, path: List[str]) -> "RouteNode":
        if len(path) == 0:
            return self

        next_path_element = path[0]
        if next_path_element in self.children:
            return self.children[next_path_element].explore(path[1::])
        elif PATH_PARAMETER in self.children:
            return self.children[PATH_PARAMETER].explore(path[1::])
        raise UnknownPathException()



from typing import Callable

from jivago.event.dispatch.message_dispatcher import MessageDispatcher
from jivago.lang.annotations import Override


class MessageDispatcherFunction(MessageDispatcher):

    def __init__(self, message_name: str, function: Callable):
        super().__init__(message_name)
        self.function = function

    @Override
    def handle(self, payload: object):
        if self._requires_payload_parameter():
            return self.function(payload)
        return self.function()

    def _requires_payload_parameter(self) -> bool:
        return self.function.__code__.co_argcount == 1



from typing import Type, List

from jivago.inject import typing_meta_helper
from jivago.lang.annotations import Override
from jivago.lang.stream import Stream
from jivago.serialization.deserialization_strategy import DeserializationStrategy, T

TYPES_WHICH_DESERIALIZE_TO_LISTS = ('List', 'Iterable', 'Collection')


class TypedListDeserializationStrategy(DeserializationStrategy):

    def __init__(self, deserializer: "Deserializer"):
        self.deserializer = deserializer

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return typing_meta_helper.is_typing_meta_collection(declared_type, TYPES_WHICH_DESERIALIZE_TO_LISTS)

    @Override
    def deserialize(self, obj: list, declared_type: Type[List[T]]) -> list:
        list_content_type = declared_type.__args__[0]
        return Stream(obj).map(lambda x: self.deserializer.deserialize(x, list_content_type)).toList()



from typing import List

from jivago.wsgi.invocation.incorrect_resource_parameters_exception import IncorrectResourceParametersException
from jivago.wsgi.invocation.route_handler_factory import RouteHandlerFactory
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class FilterChain(object):

    def __init__(self, filters: List["Filter"], route_handler_factory: RouteHandlerFactory):
        self.route_handler_factory = route_handler_factory
        self.filters = filters

    def doFilter(self, request: Request, response: Response):
        if len(self.filters) == 0:
            gotten_response = None
            for invoker in self.route_handler_factory.create_route_handlers(request):
                try:
                    gotten_response = invoker.invoke(request)
                except IncorrectResourceParametersException:
                    continue
            if gotten_response is None:
                raise IncorrectResourceParametersException()

            response.copy(gotten_response)
        else:
            self.filters[0].doFilter(request, response, self.getNextChain())

    def getNextChain(self):
        return FilterChain(self.filters[1::], self.route_handler_factory)



import unittest

from jivago.wsgi.invocation.missing_route_invocation_argument import MissingRouteInvocationArgument
from jivago.wsgi.invocation.parameter_selection.dictionary_parameter_selector import DictionaryParameterSelector
from test_utils.request_builder import RequestBuilder


class DictionaryParameterSelectorTest(unittest.TestCase):

    def setUp(self):
        self.parameter_selector = DictionaryParameterSelector()

    def test_givenStringRequestBody_whenGettingParameter_thenRaiseMissingRouteInvocationArgumentException(self):
        with self.assertRaises(MissingRouteInvocationArgument):
            self.parameter_selector.format_parameter("body", dict, RequestBuilder().build())



import unittest

from jivago.wsgi.request.headers import Headers


class HeadersTest(unittest.TestCase):

    def test_whenGettingHeaderValue_thenMatchRegardlessOfCase(self):
        headers = Headers({"Foo-Bar": "baz"})

        self.assertEqual("baz", headers['FOO_BAR'])
        self.assertEqual("baz", headers['Foo-Bar'])
        self.assertEqual("baz", headers['FOo-baR'])



from jivago.lang.stream import Stream


class Headers(object):

    def __init__(self, content=None):
        self.content = {} if content is None else content

    def __setitem__(self, key: str, value: str):
        self.content[key] = value

    def __getitem__(self, item: str) -> str:
        found_header = Stream(self.content.items()).firstMatch(
            lambda key, value: key == item or _format_camel_case(key) == _format_camel_case(item))
        if found_header:
            return found_header.get()[1]
        return None

    def values(self):
        return self.content.values()

    def keys(self):
        return self.content.keys()

    def items(self):
        return self.content.items()

    def __contains__(self, item):
        return _format_camel_case(item) in self.content


def _format_camel_case(message: str) -> str:
    return message.replace("_", "-").title()



import logging
import os
import pkgutil
import signal
from threading import Thread
from typing import List, Type, Union

import jivago
from jivago.config.abstract_context import AbstractContext
from jivago.config.debug_jivago_context import DebugJivagoContext
from jivago.config.production_jivago_context import ProductionJivagoContext
from jivago.config.properties.application_properties import ApplicationProperties
from jivago.config.properties.global_config_loader import GlobalConfigLoader
from jivago.config.properties.json_config_loader import JsonConfigLoader
from jivago.config.properties.system_environment_properties import SystemEnvironmentProperties
from jivago.config.properties.yaml_config_loader import YamlConfigLoader
from jivago.config.startup_hooks import PreInit, Init, PostInit
from jivago.lang.annotations import BackgroundWorker
from jivago.lang.registry import Registry, Annotation
from jivago.lang.stream import Stream
from jivago.scheduling.task_schedule_initializer import TaskScheduleInitializer
from jivago.scheduling.task_scheduler import TaskScheduler


class JivagoApplication(object):
    LOGGER = logging.getLogger("Jivago")

    def __init__(self, root_module=None, *, debug: bool = False,
                 context: Union[AbstractContext, Type[ProductionJivagoContext]] = None):
        self.registry = Registry()
        self.root_module = root_module

        if context is None:
            self.context = DebugJivagoContext(self.root_module, self.registry) if debug else ProductionJivagoContext(
                self.root_module, self.registry)
        elif isinstance(context, type):
            self.context = context(self.root_module, self.registry)
        else:
            self.context = context

        self.print_banner()
        self.__initialize_logger()

        self.LOGGER.info(f"Using {self.context.__class__.__name__} with root package {root_module}.")

        if self.root_module:
            self.LOGGER.info("Discovering annotated classes")
            self.__import_package_recursive(root_module)

        self.context.configure_service_locator()
        self.serviceLocator = self.context.service_locator()

        self.serviceLocator.bind(ApplicationProperties, self.__load_application_properties(self.context))
        self.serviceLocator.bind(SystemEnvironmentProperties, self.__load_system_environment_properties())

        self.LOGGER.info("Running PreInit hooks")
        self.call_startup_hook(PreInit)

        self.router = self.context.create_router_config().build(self.registry, self.serviceLocator)

        self.LOGGER.info("Running Init hooks")
        self.call_startup_hook(Init)

        self.LOGGER.info("Starting background workers")
        self.backgroundWorkers = Stream(self.get_annotated(BackgroundWorker)).map(
            lambda clazz: self.serviceLocator.get(clazz)).map(lambda worker: Thread(target=worker.run, daemon=True))
        Stream(self.backgroundWorkers).forEach(lambda thread: thread.start())

        task_schedule_initializer = TaskScheduleInitializer(self.registry, self.root_module_name)
        self.task_scheduler: TaskScheduler = self.serviceLocator.get(TaskScheduler)
        task_schedule_initializer.initialize_task_scheduler(self.task_scheduler)

        self.LOGGER.info("Running PostInit hooks")
        self.call_startup_hook(PostInit)

        signal.signal(signal.SIGTERM, self.cleanup)
        signal.signal(signal.SIGINT, self.cleanup)

    def __import_package_recursive(self, package):
        prefix = package.__name__ + "."
        for importer, modname, ispkg in pkgutil.iter_modules(package.__path__, prefix):
            module = __import__(modname, fromlist="dummy")
            if ispkg:
                self.__import_package_recursive(module)

    def __load_application_properties(self, context: AbstractContext) -> ApplicationProperties:
        composite_config_loader = GlobalConfigLoader([YamlConfigLoader(), JsonConfigLoader()])

        return Stream(context.get_config_file_locations()) \
            .firstMatch(lambda filepath: os.path.exists(filepath)) \
            .map(lambda x: composite_config_loader.read(x)) \
            .orElse(ApplicationProperties())

    def __load_system_environment_properties(self) -> SystemEnvironmentProperties:
        return SystemEnvironmentProperties(os.environ)

    def get_annotated(self, annotation: Annotation) -> List[Type]:
        return Stream(Registry().get_annotated_in_package(annotation, self.root_module_name)).map(
            lambda registration: registration.registered).toList()

    def call_startup_hook(self, hook: Annotation):
        Stream(self.get_annotated(hook)).map(lambda triggered_class: self.serviceLocator.get(triggered_class)).forEach(
            lambda x: x.run())

    def __call__(self, env, start_response):
        """wsgi entry point."""
        return self.router.route(env, start_response)

    def cleanup(self, signum, frame):
        self.LOGGER.info("Received shutdown signal. Terminating...")
        self.task_scheduler.stop()
        import sys
        sys.exit(0)

    def run_dev(self, *, port=4000, host='localhost', logging_level=logging.INFO):
        logging.getLogger().setLevel(logging_level)
        from werkzeug.serving import run_simple
        run_simple(host, port, self, processes=1, threaded=False)

    @property
    def root_module_name(self) -> str:
        return self.root_module.__name__ if self.root_module else ''

    def __initialize_logger(self):
        logging.basicConfig(
            format="%(asctime)-s [%(name)-s] [%(levelname)-s]  %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            handlers=[
                logging.StreamHandler()
            ])

    def print_banner(self):
        Stream(self.context.get_banner()).forEach(lambda x: print(x))
        print(f":: Running Jivago {jivago.__version__} ::")
        import sys
        sys.stdout.flush()
        import time
        time.sleep(0.01)



import inspect
from typing import Generic, TypeVar, Optional, Callable, Union

T = TypeVar('T')
S = TypeVar('S')


class Nullable(Generic[T]):
    """Nullable class which wraps Optional types.

    Args:
        nullable (Optional) : Item which can be None.
    """

    def __init__(self, nullable: Optional[T]):
        self._item = nullable

    def isPresent(self) -> bool:
        """Returns True if item is not None."""
        return self._item is not None

    def get(self) -> Optional[T]:
        """Gets the item if present.

        Raises:
            EmptyNullableException : Attempting to get a missing item.
            """
        if self.isPresent():
            return self._item
        raise EmptyNullableException()

    def orElse(self, default_value: T) -> T:
        """Returns the item if present, else return the supplied default value.

        Args:
            default_value : Value to return instead of a None value.
            """
        return self._item if self.isPresent() else default_value

    def orElseThrow(self, exception: Union[Exception, Callable[[], Exception]]) -> T:
        """Returns if present, raises exception if missing.

        Args:
            exception : Either an exception, or a callable which returns an exception.
        """
        if self.isPresent():
            return self._item
        if isinstance(exception, Exception):
            raise exception
        raise exception()

    def orElseFetch(self, supplier: Callable[[], T]) -> T:
        """Returns if present, invoke callable if missing.

        Args:
            supplier (Callable) : Supplied return value will be return in place of a None value. Should not require parameters.
        """
        if self.isPresent():
            return self._item
        return supplier()

    def ifPresent(self, consumer: Union[Callable[[T], None], Callable[..., None]]) -> "Nullable[T]":
        """Invoke function if value is present; otherwise does nothing.

        Args:
            consumer (Callable) : Function to be invoked with a non-nil parameter.
        """
        if self.isPresent():
            if self.__should_expand(consumer):
                consumer(*self._item)
            else:
                consumer(self._item)
        return self

    def filter(self, predicate: Union[Callable[[T], bool], Callable[..., bool]]) -> "Nullable[T]":
        """Filters item given a criterion.

        Args:
            predicate (Callable) : Invoked with a non-nil parameter. Should return a boolean.

        """
        if self.isPresent():
            if self.__should_expand(predicate):
                return self if predicate(*self._item) else Nullable.empty()
            return self if predicate(self._item) else Nullable.empty()
        return Nullable.empty()

    def map(self, callable: Union[Callable[[T], S], Callable[..., S]]) -> "Nullable[S]":
        """Maps the item when present.

        Args:
            callable (Callable) : Invoked with a non-nil parameter.
        """
        if self.isPresent():
            if self.__should_expand(callable):
                return Nullable(callable(*self._item))
            return Nullable(callable(self._item))
        return Nullable.empty()

    def __bool__(self) -> bool:
        return self.isPresent()

    def __should_expand(self, fun: Callable) -> bool:
        if inspect.isbuiltin(fun):
            return False
        if inspect.isclass(fun):
            return len(inspect.signature(fun.__init__).parameters) > 2
        sig = inspect.signature(fun)
        return len(sig.parameters) > 1

    @staticmethod
    def empty() -> "Nullable":
        return _empty


_empty = Nullable(None)


class EmptyNullableException(Exception):
    pass



import unittest

from jivago.event.config.annotations import EventHandler
from jivago.event.dispatch.jit_message_dispatcher_class import JitMessageDispatcherClass
from jivago.inject.service_locator import ServiceLocator

PAYLOAD = object()

MESSAGE_NAME = "foo"


class JitMessageDispatcherClassTest(unittest.TestCase):

    def setUp(self):
        self.service_locator = ServiceLocator()
        self.service_locator.bind(DispatcherMock, DispatcherMock)

    def test_whenHandlingMessage_thenInstantiateAndInvoke(self):
        message_dispatcher_class = JitMessageDispatcherClass(MESSAGE_NAME, DispatcherMock,
                                                             DispatcherMock.handle_message,
                                                             self.service_locator)

        result = message_dispatcher_class.handle(PAYLOAD)

        self.assertEqual(PAYLOAD, result)


class DispatcherMock(object):

    @EventHandler(MESSAGE_NAME)
    def handle_message(self, payload):
        return payload



from typing import Union, Iterable

from jivago.lang.stream import Stream
from jivago.serialization.serialization.datetime_serialization_strategy import DatetimeSerializationStrategy
from jivago.serialization.serialization_exception import SerializationException
from jivago.serialization.serialization_strategy import SerializationStrategy

BUILTIN_TYPES = (str, float, int, bool)


class Serializer(object):

    def __init__(self, additional_strategies: Iterable[SerializationStrategy] = ()):
        self.strategies = [DatetimeSerializationStrategy(), *additional_strategies]

    def serialize(self, obj: object) -> Union[dict, list]:
        if obj is None:
            return None

        if isinstance(obj, list) or isinstance(obj, tuple):
            return [self.serialize(x) for x in obj]

        if isinstance(obj, dict):
            dictionary = obj
            for key, value in dictionary.items():
                dictionary[key] = self.serialize(value)
            return dictionary
        if issubclass(type(obj), BUILTIN_TYPES):
            return obj

        if hasattr(obj, '__dict__'):
            return self.serialize({**obj.__dict__})

        return Stream(self.strategies) \
            .firstMatch(lambda s: s.can_handle_serialization(obj)) \
            .map(lambda s: s.serialize(obj)) \
            .orElseThrow(SerializationException(f"Cannot serialize type {obj.__class__}. {obj}"))



from jivago.config.router.cors_rule import CorsRule
from jivago.lang.annotations import Override
from jivago.wsgi.invocation.route_handler import RouteHandler
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class CorsRouteHandlerDecorator(RouteHandler):

    def __init__(self, cors_rule: CorsRule, route_handler: RouteHandler):
        self.cors_rule = cors_rule
        self.route_handler = route_handler

    @Override
    def invoke(self, request: Request) -> Response:
        response = self.route_handler.invoke(request)
        self.cors_rule.inject_headers(response.headers)
        return response



import unittest

from jivago.wsgi.methods import GET
from jivago.wsgi.routing.exception.unknown_path_exception import UnknownPathException
from jivago.wsgi.routing.route_registration import RouteRegistration
from jivago.wsgi.routing.table.route_node import RouteNode


class RouteNodeTest(unittest.TestCase):
    A_SIMPLE_PATH = ['hello']
    HTTP_PRIMITIVE = GET
    A_ROUTE_REGISTRATION = RouteRegistration(None, None, [""], HTTP_PRIMITIVE)
    A_LONGER_PATH = ['hello', 'goodbye']
    A_PATH_WITH_PARAMETERS = ["hello", "{id}", "delete"]
    A_PATH_WITH_DIFFERENT_PARAMETERS = ["hello", "{name}", "delete"]

    def setUp(self):
        self.rootNode = RouteNode()

    def test_whenRegisteringAPath_thenMethodInvocatorIsSavedLastChildNode(self):
        self.rootNode.register_child(self.A_SIMPLE_PATH, self.HTTP_PRIMITIVE, self.A_ROUTE_REGISTRATION)

        self.assertEqual({self.HTTP_PRIMITIVE: [self.A_ROUTE_REGISTRATION]}, self.rootNode.children['hello'].invocators)

    def test_givenPathOfMultipleElements_whenRegistering_thenMethodInvocatorIsOnlySavedAtLastChildNode(self):
        self.rootNode.register_child(self.A_LONGER_PATH, self.HTTP_PRIMITIVE, self.A_ROUTE_REGISTRATION)

        self.assertEqual({self.HTTP_PRIMITIVE: [self.A_ROUTE_REGISTRATION]},
                         self.rootNode.children['hello'].children['goodbye'].invocators)
        self.assertEqual(1, len(self.rootNode.children))
        self.assertEqual(1, len(self.rootNode.children['hello'].children))

    def test_whenExploring_thenWalkTheTreeWordByWord(self):
        self.rootNode.register_child(self.A_LONGER_PATH, self.HTTP_PRIMITIVE, self.A_ROUTE_REGISTRATION)

        node = self.rootNode.explore(self.A_LONGER_PATH)

        self.assertEqual(self.A_ROUTE_REGISTRATION, node.invocators[self.HTTP_PRIMITIVE][0])

    def test_givenUnknownPath_whenExploring_thenThrowUnknownPathException(self):
        with self.assertRaises(UnknownPathException):
            self.rootNode.explore(self.A_SIMPLE_PATH)

    def test_givenParameters_whenExploring_thenFollowPathRegardlessOfParameter(self):
        self.rootNode.register_child(self.A_PATH_WITH_PARAMETERS, self.HTTP_PRIMITIVE, self.A_ROUTE_REGISTRATION)

        node = self.rootNode.explore(self.A_PATH_WITH_PARAMETERS)
        node2 = self.rootNode.explore(self.A_PATH_WITH_DIFFERENT_PARAMETERS)

        self.assertEqual(self.A_ROUTE_REGISTRATION, node.invocators[self.HTTP_PRIMITIVE][0])
        self.assertEqual(node, node2)



import logging
from typing import List

from jivago.event.dispatch.message_dispatcher import MessageDispatcher
from jivago.event.event_bus import EventBus
from jivago.lang.annotations import Override
from jivago.lang.stream import Stream


class SynchronousEventBus(EventBus):
    LOGGER = logging.getLogger("EventBus")

    def __init__(self, message_handlers: List[MessageDispatcher]):
        self.message_handlers = message_handlers

    @Override
    def emit(self, message_name: str, payload=None):
        handlers = Stream(self.message_handlers).filter(lambda h: h.can_handle(message_name)).toTuple()

        if len(handlers) > 0:
            return Stream(handlers).map(lambda h: h.handle(payload)).filter(lambda x: x is not None).toTuple()
        else:
            self.LOGGER.warning(f"Unhandled Message {message_name}.")

    def register(self, message_handler: MessageDispatcher):
        self.message_handlers.append(message_handler)



import unittest
from unittest import mock

from jivago.inject.service_locator import ServiceLocator
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.filter_chain_factory import FilterChainFactory
from jivago.config.router.filtering.filtering_rule import FilteringRule
from jivago.wsgi.invocation.route_handler_factory import RouteHandlerFactory
from jivago.wsgi.request.request import Request

DEFAULT_FILTER: Filter = mock.create_autospec(Filter)
SPECIFIC_FILTER: Filter = mock.create_autospec(Filter)


class FilterChainFactoryTest(unittest.TestCase):

    def setUp(self):
        self.route_handler_factory_mock = mock.create_autospec(RouteHandlerFactory)
        self.filter_chain_factory = FilterChainFactory([
            FilteringRule("*", [DEFAULT_FILTER]),
            FilteringRule("/foo/bar*", [SPECIFIC_FILTER])
        ], ServiceLocator(), self.route_handler_factory_mock)

    def test_whenCreatingFilterChain_thenUseFiltersFromAllApplicableRules(self):
        filter_chain = self.filter_chain_factory.create_filter_chain(Request("GET", "/foo/bar/baz", {}, "", ""))

        self.assertEqual(2, len(filter_chain.filters))
        self.assertTrue(DEFAULT_FILTER in filter_chain.filters)
        self.assertTrue(SPECIFIC_FILTER in filter_chain.filters)



import unittest

from jivago.event.dispatch.message_dispatcher_class import MessageDispatcherClass

MESSAGE_NAME = "message_name"


class MessageDispatcherClassTest(unittest.TestCase):

    def setUp(self):
        self.instance = MessageHandlerMock()

    def test_whenHandlingEvent_thenInvokeRegisteredInstanceWithPayload(self):
        dispatcher = MessageDispatcherClass(MESSAGE_NAME, self.instance.handle_with_payload)
        payload = object()

        result = dispatcher.handle(payload)

        self.assertEqual(payload, result)

    def test_givenMissagePayloadParameter_whenHandlingEvent_thenInvokeRegisteredInstanceWithoutPayload(self):
        dispatcher = MessageDispatcherClass(MESSAGE_NAME, self.instance.handle_without_payload)

        result = dispatcher.handle(object())

        self.assertEqual("foo", result)


class MessageHandlerMock(object):

    def handle_with_payload(self, payload):
        return payload

    def handle_without_payload(self):
        return "foo"



import unittest
from unittest import mock

from jivago.inject.service_locator import ServiceLocator
from jivago.wsgi.filter.filter import Filter
from jivago.config.router.filtering.filtering_rule import FilteringRule


class FilteringRuleTest(unittest.TestCase):

    def test_givenWildcardPattern_thenMatchAnyLevelsOfDepth(self):
        rule = FilteringRule("/users/*/delete", [])

        self.assertTrue(rule.matches("/users/foo/bar/delete"))
        self.assertFalse(rule.matches("/users/foo/bar"))
        self.assertFalse(rule.matches("/users"))

    def test_givenTrailingWildcardPattern_whenMatchingRule_thenMatchForAllPathsWhichStartWithPrefix(self):
        rule = FilteringRule("/users/*", [])

        self.assertTrue(rule.matches("/users/foo/bar"))
        self.assertFalse(rule.matches("/users"))

    def test_givenSpecificPath_whenMatchingRule_thenPathMustMatchRuleExactly(self):
        rule = FilteringRule("/users", [])

        self.assertTrue(rule.matches("/users"))
        self.assertFalse(rule.matches("/users/foobar"))

    def test_givenRegexString_whenMatchingRule_thenUseRegexpAsIs(self):
        rule = FilteringRule("", [], regex_pattern=r"^/users/.*/delete$")

        self.assertTrue(rule.matches("/users/foo/delete"))
        self.assertFalse(rule.matches("/users/foo"))

    def test_givenRuleWithPeriod_whenMatchingRule_thenTreatAsLiteralPeriod(self):
        rule = FilteringRule("/users.hello", [])

        self.assertTrue(rule.matches("/users.hello"))
        self.assertFalse(rule.matches("/users_hello"))

    def test_givenSavedFilterClass_whenGettingFilters_thenGetInstanceFromServiceLocator(self):
        rule = FilteringRule("/*", [Filter])
        service_locator = ServiceLocator()
        filter_instance = object()
        service_locator.bind(Filter, filter_instance)

        filters = rule.get_filters(service_locator)

        self.assertEqual(1, len(filters))
        self.assertEqual(filter_instance, filters[0])

    def test_givenFilterInstance_whenGettingFilters_thenReturnFilterInstanceAsIs(self):
        instance = mock.create_autospec(Filter)
        rule = FilteringRule("/*", [instance])

        filters = rule.get_filters(ServiceLocator())

        self.assertEqual(1, len(filters))
        self.assertEqual(instance, filters[0])



from jivago.config.abstract_binder import AbstractBinder
from jivago.event.config.annotations import EventHandler
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Override
from jivago.lang.registry import Registry
from jivago.lang.runnable import Runnable


class RunnableEventHandlerBinder(AbstractBinder):

    def __init__(self, root_package: str, registry: Registry):
        self.root_package = root_package
        self.registry = registry

    @Override
    def bind(self, service_locator: ServiceLocator):
        for registration in self.registry.get_annotated_in_package(EventHandler, self.root_package):
            if registration.is_class_registration() and issubclass(registration.registered, Runnable):
                runner_class = registration.registered
                service_locator.bind(runner_class, runner_class)
